# include	<stdio.h>
# include	<math.h>
# include	<string.h>
# include	<time.h>
# include	<stdlib.h>
# include	<malloc.h>
# include	<ctype.h>
# include	<unistd.h>
# include	<glob.h>

# include	"nmea.h"


struct poi {
	char		*name;
	double		poi_lat;
	double		poi_long;
	double		radius;
};

struct poi_list {
	struct poi	*item;
	int		marked;
	struct poi_list	*next;
};


static struct poi_list *alist = NULL;	/* 'Avoid-point' list */
static struct poi_list *vlist = NULL;	/* 'Via-point' list */


struct poi_list *
add_to( struct poi_list *vlist, struct poi *pptr )
{
	struct poi_list	*nptr = malloc(sizeof( struct poi_list ));
	
	if (nptr != NULL) {
		nptr->item = pptr;
		nptr->marked = 0;
		nptr->next = vlist;
		vlist = nptr;
	}
	
	return vlist;
}


void
clear_marks( struct poi_list *vlist )
{

	while (vlist != NULL) {
		vlist->marked = 0;
		vlist = vlist->next;
	}

	return;
}

void
mark_in_poi_list( struct poi_list *vlist, struct poi *pptr )
{

	while (vlist != NULL) {
		if (vlist->item == pptr) {
			vlist->marked = 1;
			break;
		}
		
		vlist = vlist->next;
	}

	return;
}


int
marked_in_poi_list( struct poi_list *vlist, int test )
{

	while (vlist != NULL) {
		if (vlist->marked == test)
			return 1;
		
		vlist = vlist->next;
	}

	return 0;
}


static struct poi	*POIs = NULL;
static int		NPOI = 0;
static int 		width = 10;

static int
read_db( const char *name )
{
	char		buffer[256], *np, *latp, *longp, *proxp;
	int		n;
	struct poi	*pptr;
	FILE		*in = fopen(name, "r");
	
	if (in != NULL) {
		while (!feof(in) && fgets(buffer, sizeof buffer, in) != NULL) {
			if (buffer[0] == '#')
				/* comment */
				continue;
			
			n++;	
		}
		
		rewind(in);
		
		if (n > 0) {
			POIs = realloc(POIs, sizeof( struct poi ) * n);	/* 'POIs' will be NULL first time around: that's OK */
			pptr = &POIs[NPOI];	/* point to the first available slot in the (possibly new) array */
			
			n = 0; /* make sure the count is right this time (only correct syntax entries accepted) */
		
			while (!feof(in) && (np = fgets(buffer, sizeof buffer, in)) != NULL) {
				if (buffer[0] == '#')
					/* comment */
					continue;

				/* 'name' may be delimited by double-quotes */
				latp = np;
				
				if (np[0] == '"') {
					np++;
					
					if ((latp = strchr(np, '"')) == NULL)
						/* missing closing-quote? */
						continue;
				
					*latp++ = '\0';	/* null-terminate 'name' & point to what follows */
				}

				if ((latp = strchr(latp, ',')) == NULL)
					continue;
				
				*latp++ = '\0';	/* null-terminate 'name' & point to latitude */
				
				if ((longp = strchr(latp, ',')) == NULL)
					continue;
				
				*longp++ = '\0'; /* null-terminate 'latitude' & point to longitude */
			
				if ((proxp = strchr(longp, ',')) == NULL)
					continue;
				
				*proxp++ = '\0'; /* null-terminate 'longitude' & point to proximity */
				
				pptr->name = strdup(np);
				pptr->poi_lat = atof(latp);
				pptr->poi_long = atof(longp);
				pptr->radius = atof(proxp);
				pptr++;

				if (strlen(np) > width)
					width = strlen(np);
				
				n++;
			}
			
			NPOI += n;
		}
		
		fclose(in);
	}
	
	return (in != NULL);	/* TRUE if we read the file */
}


struct encounter {
	struct poi	*pPoi;
	double		dBest;
	struct stc	stcBest;
	struct encounter *next;
};


static struct encounter	*encounterList = NULL;


/*
 * Use Haversine formula to Calculate distance (in km) between two points specified by 
 * latitude/longitude (in numeric degrees)
 *
 */
double
distHaversine( double lat1, double lon1, double lat2, double lon2 )
{
	const double R = 6371.0; // earth's mean radius in km
	double dLat = (lat2-lat1)*M_PI/180.0;
	double dLon = (lon2-lon1)*M_PI/180.0;
	double a, c;
	
	lat1 = lat1*M_PI/180.0;
	lat2 = lat2*M_PI/180.0;

	a = sin(dLat/2.0) * sin(dLat/2.0) +
        	cos(lat1) * cos(lat2) * 
        	sin(dLon/2.0) * sin(dLon/2.0);
	c = 2.0 * atan2(sqrt(a), sqrt(1.0-a));
	
	return R * c;
}


static int
in_elist( struct encounter *eptr, struct poi *pptr )
{

	while (eptr != NULL) {
		if (eptr->pPoi == pptr)
			/* a hit! */
			return 1;
			
		eptr = eptr->next;
	}
	
	return 0;
}


static int	lflag = 0;
static int	sflag = 0;		/* stats output (requires -t) */
static int	tflag = 0;
static int	Lflag = 0;
static int	vflag = 0;

struct poi	*rptr = NULL;		/* reference point */
struct poi	*startptr = NULL;	/* start point (for time difference) */
struct poi	*endptr = NULL;		/* end point (for time difference) */

static double	*svector = NULL;	/* stats */
static int	svectorsize = 0;
static int	svectorindex = 0;


static int
approach( char *filename )
{
	char		buffer[128], adate[32];
	FILE		*in = fopen(filename, "r");
	double		diff;
	time_t		tBest;
	struct poi	*pptr;
	struct stc	stc, *r_stc = NULL, *s_stc = NULL, *e_stc = NULL;
	struct encounter *prev_eptr, *old_prev, *old_elist, *eptr, *eelist = NULL, *elist = NULL;
	
	if (in == NULL) {
		if (vflag)
			fprintf(stderr, "%s: file not readable\n", filename);
			
		return -1;
	}
	
	clear_marks(vlist);
			
	while (!feof(in) && fgets(buffer, sizeof buffer, in) != NULL) {
		if (spacetime_coords(buffer, &stc) == 0) {
			/* search the POIs for POIs that are not yet in the elist, to see if they qualify to join */
			old_elist = elist;
			old_prev = NULL;
			
			for (pptr = POIs; pptr < &POIs[NPOI]; pptr++) {
				if (!in_elist(elist, pptr)) {
					diff = distHaversine(stc.dLat, stc.dLong, pptr->poi_lat, pptr->poi_long);
			
					if (diff < pptr->radius) {
						/* a new entrant - add to elist */
						if ((eptr = malloc(sizeof( struct encounter ))) != NULL) {
							eptr->pPoi = pptr;
							eptr->dBest = diff;
							eptr->stcBest = stc;
							eptr->next = elist;
							elist = eptr;
							
							if (old_prev == NULL)
								/* this new entry's 'next' member points at the original list (old_elist) */ 
								old_prev = eptr;
						}
						
						/* if we're filtering on via points, does this one appear in the list? */
						mark_in_poi_list(vlist, pptr);
						
						/* if we're filtering on avoid points, does this one appear in the list? */
						/* TODO: make this a list! */
						mark_in_poi_list(alist, pptr);
						
						/* if we have a reference-point, does this one match? */
						if (rptr != NULL && rptr == pptr)
							r_stc = &(eptr->stcBest);
						
						/* if this is a start-point, note its best approach time */
						if (startptr != NULL && startptr == pptr)
							/* use the last encounter of the start-point in the NMEA file */
							s_stc = &(eptr->stcBest);
						
						/* if this is a end-point, note its best approach time */
						if (endptr != NULL && e_stc == NULL && endptr == pptr)
							/* use the first encounter of the end-point in the NMEA file */
							e_stc = &(eptr->stcBest);
					}
				}
			}
			
			/* search the elist (but not the new entrants) to see if they are 'improved' or 'completed' */
			eptr = old_elist;
			prev_eptr = NULL;
			
			while (eptr != NULL) {
				pptr = eptr->pPoi;
				diff = distHaversine(stc.dLat, stc.dLong, pptr->poi_lat, pptr->poi_long);
			
				if (diff < pptr->radius*2) {	/* note: hysteresis */
					/* still valid - has it improved? */
					if (diff < eptr->dBest) {
						eptr->dBest = diff;
						eptr->stcBest = stc;
					}
					
					/* move on to next one */
					prev_eptr = eptr;
					eptr = eptr->next;
				}
				else {
					struct encounter *tmp_eptr = eptr->next;
					
					/* 'eptr' is no longer a candidate - put it on the *end* of the encounterList */
					if (eelist != NULL)
						eelist->next = eptr;
					else
						encounterList = eptr;
						
					eelist = eptr;
					eptr->next = NULL;
					
					/* remove 'eptr' from  elist */
					if (prev_eptr != NULL)
						/* easy version... */
						prev_eptr->next = tmp_eptr;
					else if (old_prev != NULL)
						/* a new entry was inserted */
						old_prev->next = tmp_eptr;
					else
						/* replace the list placeholder itself */
						elist = tmp_eptr;
					
					/* move on to next one */
					eptr = tmp_eptr;
				}
			}
		}
	}

	/* we don't want to find any unmarked point in the via-list AND we don't want to find any marked point in the avoid-list */
	if (!marked_in_poi_list(vlist, 0) && !marked_in_poi_list(alist, 1)) {			
		if (utc_set && old_TZ != NULL) {
			/* I want ctime() to interpret the times as local clock times */
			setenv("TZ", old_TZ, 1);
			tzset();
			utc_set = 0;
			free(old_TZ);
			old_TZ = NULL;
		}

		if (Lflag) {
			/* list all POIs */					
			if (lflag && !vflag)
				/* just print the filename once, followed by the POI list */
				printf("%s:\n", filename);
				
			if (elist != NULL) {
				printf("Still Active:\n");
				
				for (eptr = elist; eptr != NULL; eptr = eptr->next) {
					pptr = eptr->pPoi;
					ctime_r(&(eptr->stcBest.tTime), adate);
					adate[24] = '\0';
						
					if (vflag && lflag)
						/* print the filename with each item of the POI list */
						printf("%s: ", filename);
					
					if (eptr->dBest >= 0.2)			
						printf("%*s: %s (closest approach: %.3fkm)\n", width, pptr->name, &adate[4], eptr->dBest);
					else
						printf("%*s: %s (closest approach: %3dm)\n", width, pptr->name, &adate[4], (int)(eptr->dBest*1000.0 + 0.5));
				}
				
				printf("\n");
			}
			
			for (eptr = encounterList; eptr != NULL; eptr = eptr->next) {
				struct stc *ref = &encounterList->stcBest;
				int	pIndex;
				
				if (rptr != NULL && r_stc != NULL)
					/* make the timings relative to the reference-point */
					ref = r_stc;
					
				pptr = eptr->pPoi;
				pIndex = pptr - &POIs[0];
				ctime_r(&(eptr->stcBest.tTime), adate);
				adate[24] = '\0';		
						
				if (vflag && lflag)
					/* print the filename with each item of the POI list */
					printf("%s: ", filename);
					
				if (eptr->dBest >= 0.2)			
					printf("%*s(#%d): %s (range: %.3fkm at %.1fmin)\n", width, pptr->name, pIndex, &adate[4], eptr->dBest, (eptr->stcBest.tTime - ref->tTime)/60.0);
				else
					printf("%*s(#%d): %s (range: %3dm at %.1fmin)\n", width, pptr->name, pIndex, &adate[4], ( int )(eptr->dBest*1000.0 + 0.5), (eptr->stcBest.tTime - ref->tTime)/60.0);
			}
		}
		
		if (tflag && (s_stc != NULL && e_stc != NULL)) {
			/* time taken between two POIs */					
			if (lflag)
				/* just print the filename once, followed by the info */
				printf("%s:\n", filename);
				
			diff = (e_stc->tTime - s_stc->tTime)/60.0;
			
			if (diff < 0.0)
				diff = -diff;
					
			printf("%*s - %*s: %.1fmin)\n", width, startptr->name, width, endptr->name, diff);
			
			if (sflag) {
				if (svectorindex >= svectorsize) {
					/* I need more space */
					svectorsize += 10;
					svector = realloc(svector, sizeof( double ) * svectorsize);
				}
				
				svector[svectorindex++] = diff;
			}
		}
	}
	
	/* free the elist */
	for (eptr = elist; eptr != NULL; eptr = eptr->next)
		free(eptr);
		
	eptr = NULL;
	
	/* free the encounterList */
	for (eptr = encounterList; eptr != NULL; eptr = eptr->next)
		free(eptr);
		
	encounterList = NULL;
	fclose(in);
	return 0;
}


static int
cmpfnc( const void *a, const void *b )
{

	return ( int )(*(( double * ) a) - *(( double * ) b));
}


static void
stats( void )
{
	double	avg = 0.0, sd = 0.0;
	int	i;
	
	if (svectorindex > 0) {
		for (i = 0; i < svectorindex; i++)
			avg += svector[i];
	
		avg /= svectorindex;
		
		for (i = 0; i < svectorindex; i++) {
			double d = svector[i] - avg;
	
			sd += d*d;
		}
	
		sd = sqrt(sd / svectorindex);
		printf("Average time: %.3fmin, S.D: %.3f\n", avg, sd);
		
		qsort(svector, svectorindex, sizeof( double ), cmpfnc);
		
		if (svectorindex >= 3) {
			/* leave off the highest and lowest scores */
			avg = 0.0;
		
			for (i = 1; i < svectorindex-1; i++)
				avg += svector[i];
	
			avg /= svectorindex - 2;
			sd = 0.0;
		
			for (i = 1; i < svectorindex-1; i++) {
				double d = svector[i] - avg;
	
				sd += d*d;
			}
	
			sd = sqrt(sd / (svectorindex - 2));
			printf("Figure-skating style Average time: %.3fmin, S.D: %.3f\n", avg, sd);
		}
		
		printf("Median time: %.3fmin\n", svector[svectorindex/2]);
	}
	
	return;
}
				

int
main (int argc, char **argv)
{
	char		*cvalue = NULL;
	int		index, explicit = 0;
	int		c;
	struct poi	*pptr;
	glob_t		gblock;
	
	glob("approach.poi", GLOB_NOSORT, NULL, &gblock);	
	glob("~/.approach/*.poi", GLOB_NOSORT | GLOB_APPEND | GLOB_TILDE, NULL, &gblock);	
	glob("/usr/share/approach/*.poi", GLOB_NOSORT | GLOB_APPEND | GLOB_TILDE, NULL, &gblock);
	
	for (c = 0; c < gblock.gl_pathc; c++)
		read_db(gblock.gl_pathv[c]);

	globfree(&gblock);
	
	opterr = 0;
	
	while ((c = getopt(argc, argv, "hlsvR:V:A:t:L")) != -1) {
		int	via;
		
		switch (c) {
		case 'l':
			lflag = 1;
			break;
			
		case 's':
			sflag = 1;
			break;
			
		case 'V':
			/* filter output based on whether the route goes *V*ia all of a set of points */
			via = atoi(optarg);
			
			if (via < 0 || via >= NPOI) {
				fprintf(stderr, "Invalid via-point: %d\n", via);
				return -1;
			}
			
			vlist = add_to(vlist, &POIs[via]);
			break;
			
		case 'A':
			/* filter output based on whether the route *A*voids any one of a set of points */
			via = atoi(optarg);
			
			if (via < 0 || via >= NPOI) {
				fprintf(stderr, "Invalid avoid-point: %d\n", via);
				return -1;
			}
			
			alist = add_to(alist, &POIs[via]);
			break;
			
		case 'R':
			/* filter output based on whether the route goes via some *R*eference point */
			via = atoi(optarg);
			
			if (via < 0 || via >= NPOI) {
				fprintf(stderr, "Invalid reference-point: %d\n", via);
				return -1;
			}
			
			rptr = &POIs[via];
			vlist = add_to(vlist, rptr); /* reference point is also a via-point */
			Lflag = 1;	/* 'R' rather implies 'L' */
			break;
			
		case 't':
			/* print time between some pair of points */
			via = atoi(optarg);
			
			if (via < 0 || via >= NPOI) {
				fprintf(stderr, "Invalid via-point: %d\n", via);
				return -1;
			}
			
			startptr = &POIs[via];
			vlist = add_to(vlist, startptr); /* start point is also a via-point */
			optarg = argv[optind++];	/* second arg */
			via = atoi(optarg);
			
			if (via < 0 || via >= NPOI) {
				fprintf(stderr, "Invalid via-point: %d\n", via);
				return -1;
			}
			
			endptr = &POIs[via];
			vlist = add_to(vlist, endptr); /* end point is also a via-point */
			tflag = 1;
			explicit = 1;
			break;
			
		case 'L':
			Lflag = 1;
			explicit = 1;
			break;
			
		case 'v':
			vflag = 1;
			break;
		
		case 'h':
			fprintf(stderr, "Usage:\n");
			fprintf(stderr, "   %s\n", argv[0]);
			fprintf(stderr, "or %s <options> <NMEA files>\n\n", argv[0]);
			fprintf(stderr, "    <options> ::= <option> | <option> <options>\n\n");
			fprintf(stderr, "    <option> ::= <unique_option> | <multi_option>\n\n");
			fprintf(stderr, "    <multi_option> ::=    /* each <multi_option> can appear several times */\n");
			fprintf(stderr, "      \"-A\" <poi>          /* add an Avoid point */\n");
			fprintf(stderr, "      \"-V\" <poi>          /* add a Via point */\n\n");
			fprintf(stderr, "    <unique_option> ::=   /* each <unique_option> can appear only once */\n");
			fprintf(stderr, "      \"-R\" <poi>          /* set a Reference point */\n");
			fprintf(stderr, "      \"-t\" <poi> <poi>    /* print time elapsed between a pair of points */\n");
			fprintf(stderr, "      \"-s\"                /* print stats of times reported by '-t' option */\n");
			fprintf(stderr, "      \"-v\"                /* increase verbosity */\n");
			fprintf(stderr, "      \"-L\"                /* print a List of all POIs visited */\n");
			fprintf(stderr, "      \"-h\"                /* this help message! */\n");
			fprintf(stderr, "      \"-l\"                /* prefix NMEA filename to all printout lines */\n\n");
			fprintf(stderr, "    <poi> ::= <integer>   /* run with no options to get a POI list */\n");
			return 0;
			
		case '?':
			if (isprint (optopt))
				fprintf(stderr, "Unknown option `-%c'.\n", optopt);
			else
				fprintf(stderr, "Unknown option character `\\x%x'.\n", optopt);
			
			return 1;
			
		default:
			abort();
		}
	}
	
	if (sflag) {
		if (startptr == NULL || endptr == NULL)
			/* can't do stats without a timing-request */
			sflag = 0;
		else {
			/* start the vector off with some space at least */
			svector = calloc(sizeof( double ), 10);
			svectorsize = 10;
			svectorindex = 0;
		}
	}
	
	if (!explicit)
		/* no explicit command given: assume 'L' (List) */
		Lflag = 1;
	
	if (optind < argc-1)
		/* multiple filename arguments: force lflag on */
		lflag = 1;

	if (optind < argc) {		
		for (index = optind; index < argc; index++)
			approach(argv[index]);
		
		if (sflag)
			stats();
	}
	else {
		struct poi	*pptr;
		
		/* list the POIs */
		for (c = 0, pptr = POIs; pptr < &POIs[NPOI]; pptr++, c++)
			printf("POI #%d: %s\n", c, pptr->name);
	}
		
	return 0;
}
