# include	<stdio.h>
# include	<time.h>
# include	<sys/types.h>
# include	<sys/stat.h>
# include	<unistd.h>
# include	<utime.h>
# include	<stdlib.h>
# include	<ctype.h>

# include	"nmea.h"


int
main( int argc, char *argv[] )
{
	char		buffer[256];
	struct stat	statbuf;
	int		i, j;
	int		hour, min, sec, nflag = 0, index, c;
	time_t		tt;
	struct tm	tm_then;
	struct utimbuf	ut;
	FILE		*in;
	
	while ((c = getopt(argc, argv, "n")) != -1) {
		int	via;
		
		switch (c) {
		case 'n':
			nflag = 1;	/* read NMEA file to find date */
			break;
			
		case 'V':
			via = atoi(optarg);
			break;
						
		case '?':
			if (isprint(optopt))
				fprintf(stderr, "Unknown option `-%c'.\n", optopt);
			else
				fprintf(stderr, "Unknown option character `\\x%x'.\n", optopt);
			
			return 1;
			
		default:
			abort();
		}
	}

	if (nflag) {
		if (optind < argc) {		
			for (index = optind; index < argc; index++) {
				if ((in = fopen(argv[index], "r")) != NULL) {
					while (1) {
						struct stc stc;
						
						fgets(buffer, sizeof buffer, in);
						
						if (spacetime_coords(buffer, &stc) == 0) {
							/* tTime is in UTC */
							printf("New datestamp for %s: %s", argv[index], ctime(&stc.tTime));
							ut.actime = stc.tTime;
							ut.modtime = stc.tTime;
							utime(argv[index], &ut);	
							break;
						}
					}
					
					fclose(in);
				}
			}
		}
	}	
	else {
	    if (argc < 4) {
		fprintf(stderr, "Usage: %s [[[YYYY] MM] DD] HH MM <filename>\n", argv[0]);
		return 1;
	    }
	    
	    if (stat(argv[argc-1], &statbuf) == 0 && S_ISREG(statbuf.st_mode)) {
		/* set time */
		localtime_r(&statbuf.st_mtime, &tm_then);
		
		for (i = argc-2, j = 0; i >= optind; j++, i--) {
			//printf("%s\n", argv[i]);
			
			switch (j) {
			case 0:	/* minutes */
				min = tm_then.tm_min = atoi(argv[i]);
				sec = tm_then.tm_sec = 0;
				break;
				
			case 1: /* hours */
				hour = tm_then.tm_hour = atoi(argv[i]);
				break;
				
			case 2: /* day */
				tm_then.tm_mday = atoi(argv[i]);
				break;
				
			case 3: /* month */
				tm_then.tm_mon = atoi(argv[i]) - 1;
				break;
				
			case 4: /* year */
				tm_then.tm_year = atoi(argv[i]) - 1900;
				break;
				
			default:
				fprintf(stderr, "Date not understood\n");
				return 2;
			}
		}
		
		tm_then.tm_isdst = 0;	/* start off assuming DST isn't on */	
		tt = mktime(&tm_then);
		
		/* first call to mktime will set 'isdst' correctly - so reset the HH:MM:SS part */
		tm_then.tm_hour = hour;
		tm_then.tm_min = min;
		tm_then.tm_sec = sec;
		tt = mktime(&tm_then);
		printf("New datestamp: (%s) %s", tm_then.tm_isdst? "BST": "GMT", ctime(&tt));
		
		ut.actime = tt;
		ut.modtime = tt;
		utime(argv[argc-1], &ut);
	    }	
	    else {
		fprintf(stderr, "Not a regular file: %s\n", argv[argc-1]);
		return 3;
	    }
	}

	return 0;
}
