# include	<stdio.h>
# include	<string.h>
# include	<time.h>
# include	<stdlib.h>

# include	"nmea.h"


int		utc_set = 0;
char		*old_TZ = NULL;


static int
dec2( char *str )
{
	
	return (int)(str[0]-'0')*10 + (int)(str[1]-'0');
}


static double
lat_conv( char *str )
{
	char	*cp = strchr(str, ',');
	int	deg = dec2(str);
	double	ret = atof(&str[2])/60.0 + deg;
	
	if (cp != NULL && cp[1] == 'S')
		ret = -ret;
	
	return ret;
}


static int
dec3( char *str )
{
	
	return (int)(str[0]-'0')*100 + dec2(&str[1]);
}


static double
long_conv( char *str )
{
	char	*cp = strchr(str, ',');
	int	deg = dec3(str);
	double	ret = atof(&str[3])/60.0 + deg;
	
	if (cp != NULL && cp[1] == 'W')
		ret = -ret;
	
	return ret;
}


int
spacetime_coords( char *buffer, struct stc *stc )
{
	char		*fixtime, *fixlat, *fixlong, *speed, *bearing, *fixdate, *cp;
	struct tm	tms;

	/* warning: this routine tramples all over 'buffer' */
	if (stc != NULL && strncmp(buffer, "$GPRMC", 6) == 0) {
		/* comma-separated */
		if ((fixtime = strchr(buffer, ',')) == NULL)
			/* missing comma? */
			return -1;
			
		*fixtime++ = '\0';
			
		if ((cp = strchr(fixtime, ',')) == NULL)
			/* missing comma? */
			return -1;
			
		*cp++ = '\0';
			
		if (*cp != 'A')
			/* not an active fix? */
			return -1;
			
		if ((fixlat = strchr(cp, ',')) == NULL)
			/* missing comma? */
			return -1;
			
		*fixlat++ = '\0';
			
		if ((cp = strchr(fixlat, ',')) == NULL)
			/* missing comma? */
			return -1;
		
		/* leave the comma and 'N'/'S' following the lat */
		cp++;
			
		if ((fixlong = strchr(cp, ',')) == NULL)
			/* missing comma? */
			return -1;
			
		*fixlong++ = '\0';
			
		if ((cp = strchr(fixlong, ',')) == NULL)
			/* missing comma? */
			return -1;
		
		/* leave the comma and 'E'/'W' following the long */
		cp++;
			
		if ((speed = strchr(cp, ',')) == NULL)
			/* missing comma? */
			return -1;
			
		*speed++ = '\0';
			
		if ((bearing = strchr(speed, ',')) == NULL)
			/* missing comma? */
			return -1;
			
		*bearing++ = '\0';
			
		if ((fixdate = strchr(bearing, ',')) == NULL)
			/* missing comma? */
			return -1;
			
		*fixdate++ = '\0';
			
		if ((cp = strchr(fixdate, ',')) == NULL)
			/* missing comma? */
			return -1;
		
		/* null-terminate the date, don't care about the rest */
		*cp++ = '\0';
		
		tms.tm_sec = dec2(&fixtime[4]);
		tms.tm_min = dec2(&fixtime[2]);
		tms.tm_hour = dec2(&fixtime[0]);
		tms.tm_mday = dec2(&fixdate[0]);
		tms.tm_mon = dec2(&fixdate[2]) - 1;
		tms.tm_year = dec2(&fixdate[4]) + 100;
		tms.tm_isdst = 0; /* no daylight-saving system used with UTC! */
		
		if (!utc_set) {
			char	*tzp = getenv("TZ");
			
			/* I want mktime() to interpret the times as UTC */
			if (tzp != NULL) {
				old_TZ = strdup(tzp);
				setenv("TZ", "", 1);
				tzset();
				utc_set = 1;
			}
		}

		stc->tTime = mktime(&tms);
		
		stc->dLat = lat_conv(fixlat);
		stc->dLong = long_conv(fixlong);
		stc->dAlt = 0;
		return 0;
	}
	
	return -1;
}
